from .base_control import ResourceLister, Describer, MultiLister, NoResults, GenericDescriber, DialogFieldResourceListSelector, DeleteResourceDialog, SingleRelationLister
from .common import Common, SessionAwareDialog, BaseChart
from .termui.dialog import DialogControl, DialogFieldText, DialogFieldLabel, DialogFieldButton, DialogFieldCheckbox
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes
from .ssh import SSHList
import subprocess
from pathlib import Path
import json
import jq
import botocore
import time
import datetime
from .arn import ARN

class SGResourceLister(ResourceLister):
  prefix = 'sg_list'
  title = 'Security Groups'
  command_palette = ['sg', 'securitygroup']

  def __init__(self, *args, **kwargs):
    self.resource_key = 'ec2'
    self.list_method = 'describe_security_groups'
    self.item_path = '.SecurityGroups'
    self.column_paths = {
      'group id': '.GroupId',
      'name': '.GroupName',
      'vpc': '.VpcId',
      'ingress rules': self.determine_ingress_rules,
      'egress rules': self.determine_egress_rules,
    }
    self.imported_column_sizes = {
      'group id': 30,
      'name': 30,
      'vpc': 15,
      'ingress rules': 20,
      'egress rules': 20,
    }
    self.describe_command = SGDescriber.opener
    self.open_command = SGRelated.opener
    self.open_selection_arg = 'compare_value'

    self.imported_column_order = ['group id', 'name', 'vpc', 'ingress rules', 'egress rules']
    self.sort_column = 'name'
    self.primary_key = 'group id'
    super().__init__(*args, **kwargs)
    self.add_hotkey('i', self.ingress, 'View ingress rules')
    self.add_hotkey('e', self.egress, 'View egress rules')

  def determine_ingress_rules(self, sg):
    return len(sg['IpPermissions'])

  def determine_egress_rules(self, sg):
    return len(sg['IpPermissionsEgress'])

  def ingress(self, *args):
    if self.selection is not None:
      Common.Session.push_frame(SGRuleLister.opener(sg_entry=self.selection))

  def egress(self, *args):
    if self.selection is not None:
      Common.Session.push_frame(SGRuleLister.opener(sg_entry=self.selection, egress=True))

class SGRuleLister(ResourceLister):
  prefix = 'sg_rule_list'
  title = 'SG Rules'
  well_known = {
    20: 'ftp',
    22: 'ssh',
    23: 'telnet',
    25: 'smtp',
    80: 'http',
    110: 'pop3',
    220: 'imap',
    443: 'https',
    989: 'ftps',
    1437: 'dashboard-agent',
    1443: 'mssql',
    3306: 'mysql',
    3389: 'rdp',
    5432: 'pgsql',
    6379: 'redis',
    8983: 'solr',
    9200: 'es',
  }

  def title_info(self):
    return '{0}: {1}'.format('Egress' if self.egress else 'Ingress', self.sg_entry['group id'])

  def __init__(self, parent, alignment, dimensions, sg_entry=None, egress=False, *args, **kwargs):
    self.resource_key = 'ec2'
    self.list_method = 'describe_security_groups'
    if sg_entry is None:
      raise ValueError('sg_entry is required')
    self.sg_entry = sg_entry
    self.egress = egress
    self.list_kwargs = {'GroupIds': [self.sg_entry['group id']]}
    self.item_path = '.SecurityGroups[0].IpPermissions{0}'.format('Egress' if egress else '')
    self.column_paths = {
      'protocol': self.determine_protocol,
      'name': self.determine_name,
      'ip addresses': self.determine_ips,
      'security groups': self.determine_sgs,
      'port range': self.determine_port_range,
    }
    self.imported_column_sizes = {
      'name': 10,
      'protocol': 5,
      'ip addresses': 40,
      'security groups': 50,
      'port range': 20,
    }
    self.imported_column_order = ['name', 'protocol', 'ip addresses', 'security groups', 'port range']
    self.sort_column = 'port range'
    super().__init__(parent, alignment, dimensions, *args, **kwargs)

  def determine_name(self, rule):
    if rule['IpProtocol'] == '-1':
      return '<all>'
    if rule['FromPort'] != rule['ToPort']:
      return ''
    if rule['FromPort'] not in SGRuleLister.well_known:
      return ''
    return SGRuleLister.well_known[rule['FromPort']]

  def determine_protocol(self, rule):
    if rule['IpProtocol'] == '-1' or rule['IpProtocol'] == -1:
      return '<all>'
    return rule['IpProtocol']

  def determine_ips(self, rule):
    if 'IpRanges' not in rule:
      return ''
    return ','.join(cidrs['CidrIp'] for cidrs in rule['IpRanges'])

  def determine_sgs(self, rule):
    if 'UserIdGroupPairs' not in rule:
      return ''
    return ','.join(pair['GroupId'] for pair in rule['UserIdGroupPairs'])

  def determine_port_range(self, rule):
    if rule['IpProtocol'] == '-1':
      return '0-65535'
    if rule['FromPort'] != rule['ToPort']:
      return '{0}-{1}'.format(rule['FromPort'], rule['ToPort'])
    return str(rule['FromPort'])

class SGDescriber(Describer):
  prefix = 'sg_browser'
  title = 'Security Group'

  def __init__(self, parent, alignment, dimensions, entry, *args, entry_key='group id', **kwargs):
    self.resource_key = 'ec2'
    self.describe_method = 'describe_security_groups'
    self.describe_kwarg_name = 'GroupIds'
    self.describe_kwarg_is_list = True
    self.object_path='.SecurityGroups[0]'
    super().__init__(parent, alignment, dimensions, *args, entry=entry, entry_key=entry_key, **kwargs)

class SGRelated(MultiLister):
  prefix = 'sg_related'
  title = 'Resources using Security Group'

  def title_info(self):
    return self.compare_value

  def __init__(self, *args, **kwargs):
    kwargs['compare_key'] = 'group id'
    self.resource_descriptors = [
      {
        'resource_key': 'ec2',
        'list_method': 'describe_instances',
        'list_kwargs': {},
        'item_path': '[.Reservations[].Instances[]]',
        'column_paths': {
          'type': lambda x: 'EC2 Instance',
          'id': '.InstanceId',
          'name': self.tag_finder_generator('Name'),
        },
        'hidden_columns': {},
        'compare_as_list': True,
        'compare_path': '[.SecurityGroups[].GroupId]',
      },
      {
        'resource_key': 'rds',
        'list_method': 'describe_db_instances',
        'list_kwargs': {},
        'item_path': '.DBInstances',
        'column_paths': {
          'type': lambda x: 'RDS Instance',
          'id': '.DBInstanceIdentifier',
          'name': '.Endpoint.Address',
        },
        'hidden_columns': {},
        'compare_as_list': True,
        'compare_path': '[.VpcSecurityGroups[].VpcSecurityGroupId]',
      },
    ]
    super().__init__(*args, **kwargs)



import curses
from blessed import Terminal
import termios
import tty
import signal
import time
from .alignment import TopLeftAnchor, Dimension, CenterAnchor
from .common import Commons
from .control import Control, Border, BorderStyleContinuous
from .list_control import ListControl, ListEntry
from .dialog import DialogControl, DialogFieldLabel, DialogFieldText
from .color import ColorGold, ColorGreen, ColorMagenta
from .block import Block
import numpy as np
from .screen import Screen
import shutil
import tempfile
from pathlib import Path
import threading

FrameRate = 0.02

class ControlCodes:
  A = '\x01'
  B = '\x02'
  C = '\x03'
  D = '\x04'
  E = '\x05'
  F = '\x06'
  G = '\x07'
  K = '\x0b'
  L = '\x0c'
  M = '\x0d'
  N = '\x0e'
  O = '\x0f'
  P = '\x10'
  Q = '\x11'
  R = '\x12'
  S = '\x13'
  T = '\x14'
  U = '\x15'
  V = '\x16'
  W = '\x17'
  X = '\x18'
  Y = '\x19'
  Z = '\x1a'

class KeyDisplay(Control):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.text = '<no key>'

  def input(self, key):
    if key.is_sequence:
      self.text = '{0}'.format((str(key), key.name, key.code))
    else:
      self.text = key

  def paint(self):
    super().paint()
    tl = self.topleft()
    Commons.UIInstance.print(self.text, xy=tl)

class UI:
  def __init__(self):
    if Commons.UIInstance is not None:
      raise RuntimeError('UI is a singleton.')
    Commons.UIInstance = self
    self.debug = 1
    self.clear_log()
    self.top_block = Block(None, TopLeftAnchor(0, 0), Dimension('100%', '100%'), tag='top')
    self.last_paint = time.time()
    self.last_size_refresh = time.time()
    self.term = Terminal()
    self._w = self.term.width
    self._h = self.term.height
    self.exit = False
    signal.signal(signal.SIGINT, self.set_exit)
    self.buf = Screen(self)
    self.tickers = []
    self.cache_dir = None
    self.mutex = threading.Lock()
    self.input_buffer = []
    self.dirty = False

  def clear_log(self):
    if self.debug == 0:
      return
    with open('log.log', 'w') as f:
      pass

  def log(self, line, level=1):
    if self.debug < level:
      return
    with open('log.log', 'a') as f:
      f.write(line)
      f.write('\n')

  def filecache(self, obj, cb, *args, **kwargs):
    if self.cache_dir is None:
      self.cache_dir = tempfile.mkdtemp(prefix='awsc')
    objpath = Path(self.cache_dir) / obj
    if not objpath.exists():
      data = cb(str(objpath.resolve()), *args, **kwargs)
      return data
    try:
      with objpath.open('r') as f:
        return f.read()
    except UnicodeDecodeError:
      with objpath.open('rb') as f:
        return f.read()

  def print(self, out, xy=None, color=None, wrap=False, bounds=None, bold=False):
    self.dirty = True
    if bounds is None:
      bounds = ((0, self.w), (0, self.h))
    if xy is None:
      xy = (bounds[0][0], bounds[1][0])
    end = False
    if (xy[0] > bounds[0][1] and not wrap) or xy[1] > bounds[1][1]:
      return
    while not end:
      space = bounds[0][1] - xy[0]
      if len(out) > space:
        part = out[:space]
        out = out[space:]
      else:
        space = len(out)
        part = out
        out = ''
        end = True
      p = 0
      for i in range(xy[0], xy[0]+space):
        try:
          char = self.buf[xy[1]][i]
          char.value = part[p]
          char.color = color
          char.bold = bold
          char.dirty = True
        except IndexError:
          break
        p += 1
      xy = (bounds[0][0], xy[1] + 1)
      if not wrap or xy[1] >= bounds[1][1]:
        end = True

  def refresh_size(self):
    if self.last_paint - self.last_size_refresh > 1:
      self.last_size_refresh = self.last_paint
      self._w = self.term.width
      self._h = self.term.height

  @property
  def w(self):
    self.refresh_size()
    return self._w

  @property
  def h(self):
    self.refresh_size()
    return self._h

  @property
  def dim(self):
    self.refresh_size()
    return (self._w, self._h)

  def set_exit(self, *args, **kwargs):
    self.exit = True

  def before_paint(self):
    self.top_block.before_paint()

  def progress_bar_paint(self, perc):
    width = int(self.w * perc)
    with self.term.location(0, self.h - 1):
      perc_t = '{0:.2f}% '.format(perc * 100)
      print('\033[38;5;220m' + perc_t + ((width - len(perc_t)) * '░') + '\033[0m', end='')

  def check_one_key(self):
    return self.term.inkey(0.01, 0.01)

  def paint(self):
    self.buf.clear()
    self.top_block.paint()
    with self.term.location(0, 0):
      self.buf.output()
    self.last_paint = time.time()
    self.dirty = False

  def unraw(self, cmd, *args, **kwargs):
    termios.tcsetattr(self.term._keyboard_fd, termios.TCSAFLUSH, self.restore)
    self.term._line_buffered = True
    self.term.stream.write(self.term.normal_cursor)
    self.term.stream.flush()
    self.term.stream.write(self.term.exit_fullscreen)
    self.term.stream.flush()
    try:
      return cmd(*args, **kwargs)
    finally:
      self.term.stream.write(self.term.enter_fullscreen)
      self.term.stream.flush()
      self.term.stream.write(self.term.hide_cursor)
      self.term.stream.flush()
      tty.setraw(self.term. _keyboard_fd, termios.TCSANOW)
      self.term._line_buffered = False

  def buffer_input(self):
    while True:
      key = self.term.inkey(None, 0.01)
      if key is not None and key != u'':
        self.mutex.acquire()
        try:
          self.input_buffer.append(key)
        finally:
          self.mutex.release()

  def process_input_buffer(self, st):
    self.mutex.acquire()
    try:
      if len(self.input_buffer) > 0:
        key = self.input_buffer[0]
        self.input_buffer = self.input_buffer[1:]
        if key == '\x03':
          raise KeyboardInterrupt
        self.top_block.input(key)
        ct = time.time()
        if ct - st > FrameRate - 0.02:
          return
    finally:
      self.mutex.release()

  def main(self):
    global FrameRate
    self.restore = termios.tcgetattr(self.term._keyboard_fd)
    try:
      with self.term.fullscreen():
        with self.term.hidden_cursor():
          with self.term.raw():
            t = threading.Thread(target=self.buffer_input, daemon=True)
            t.start()
            while not self.exit:
              #print(self.term.clear())
              st = time.time()
              self.process_input_buffer(st)
              for func in self.tickers:
                func()
              self.before_paint()
              if self.dirty or time.time() - self.last_paint > 3:
                self.paint()
              t2 = time.time()
              #self.log('Frame time: {0:.2f}ms'.format((t2 - st) * 1000), level=1)
              d = FrameRate - (t2 - st)
              if d > 0:
                time.sleep(d)
    except KeyboardInterrupt:
      pass
    finally:
      if self.cache_dir is not None:
        shutil.rmtree(self.cache_dir)



import re
from .common import Common, SessionAwareDialog
from .termui.alignment import TopLeftAnchor, TopRightAnchor, CenterAnchor, Dimension
from .termui.dialog import DialogControl, DialogFieldLabel, DialogFieldText
from .termui.list_control import ListControl, ListEntry
from .termui.common import Commons
from .termui.control import Control, Border
from .info import HotkeyDisplay
from botocore import exceptions

class DeleteContextDialog(SessionAwareDialog):
  def __init__(self, parent, alignment, dimensions, name='', caller=None, *args, **kwargs):
    kwargs['ok_action'] = self.accept_and_close
    kwargs['cancel_action'] = self.close
    kwargs['border'] = Border(Common.border('default'), Common.color('modal_dialog_border'), 'Delete Context', Common.color('modal_dialog_border_title'))
    super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
    self.name = name
    self.add_field(DialogFieldLabel([('Delete context "', Common.color('modal_dialog_label')), (name, Common.color('modal_dialog_label_highlight')), ('"?', Common.color('modal_dialog_label'))]))
    self.highlighted = 1
    self.caller = caller

  def input(self, inkey):
    if inkey.is_sequence and inkey.name == 'KEY_ESCAPE':
      self.close()
      return True
    return super().input(inkey)

  def accept_and_close(self):
    Common.Configuration.delete_context(self.name)
    self.close()

  def close(self):
    if self.caller is not None:
      self.caller.reload_contexts()
    super().close()

class ImportContextDialog(SessionAwareDialog):
  def __init__(self, parent, alignment, dimensions, caller=None, *args, **kwargs):
    self.accepts_inputs = True
    kwargs['border'] = Border(Common.border('default'), Common.color('modal_dialog_border'), 'New Context', Common.color('modal_dialog_border_title'))
    kwargs['ok_action'] = self.accept_and_close
    kwargs['cancel_action'] = self.close
    super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
    self.caller = caller
    self.add_field(DialogFieldLabel('Enter AWS context details'))
    self.error_label = DialogFieldLabel('', default_color=Common.color('modal_dialog_error'))
    self.add_field(self.error_label)
    self.add_field(DialogFieldLabel(''))
    self.name_field = DialogFieldText('Name:', label_min=12, color=Common.color('context_add_modal_dialog_textfield', 'modal_dialog_textfield'), selected_color=Common.color('context_add_modal_dialog_textfield_selected', 'modal_dialog_textfield_selected'), label_color=Common.color('context_add_modal_dialog_textfield_label', 'modal_dialog_textfield_label'))
    self.add_field(self.name_field)
    creds = Common.Session.service_provider.env_session().get_credentials()
    if creds is None or not creds.access_key or not creds.secret_key:
      Common.Session.set_message('No valid credentials in environment.', Common.color('message_error'))
      self.close()
      return
    self.access_key = creds.access_key
    self.secret_key = creds.secret_key
    self.access_key_field = DialogFieldLabel([('Access key: ', Common.color('generic')), (creds.access_key, Common.color('highlight'))], centered=False)
    self.add_field(self.access_key_field)
    self.secret_key_field = DialogFieldLabel([('Secret key: ', Common.color('generic')), ('*' * len(creds.secret_key), Common.color('highlight'))], centered=False)
    self.add_field(self.secret_key_field)

  def input(self, inkey):
    if not self.accepts_inputs:
      return True
    if inkey.is_sequence and inkey.name == 'KEY_ESCAPE':
      self.close()
      return True
    return super().input(inkey)

  def accept_and_close(self):
    if self.name_field.text == '':
      self.error_label.text = 'Name cannot be blank.'
      return
    try:
      sts = Common.Session.service_provider.whoami(keys={'access': self.access_key, 'secret': self.secret_key})
    except exceptions.ClientError as e:
      self.error_label.text = 'Key verification failed'
      return

    self.accepts_inputs = False

    Common.Configuration.add_or_edit_context(self.name_field.text, sts['Account'], self.access_key, self.secret_key)
    self.close()

  def close(self):
    if self.caller is not None:
      self.caller.reload_contexts()
    super().close()

class AddContextDialog(SessionAwareDialog):
  def __init__(self, parent, alignment, dimensions, caller=None, *args, **kwargs):
    self.accepts_inputs = True
    kwargs['border'] = Border(Common.border('default'), Common.color('modal_dialog_border'), 'New Context', Common.color('modal_dialog_border_title'))
    kwargs['ok_action'] = self.accept_and_close
    kwargs['cancel_action'] = self.close
    super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
    self.add_field(DialogFieldLabel('Enter AWS context details'))
    self.error_label = DialogFieldLabel('', default_color=Common.color('modal_dialog_error'))
    self.add_field(self.error_label)
    self.add_field(DialogFieldLabel(''))
    self.name_field = DialogFieldText('Name:', label_min=16, color=Common.color('context_add_modal_dialog_textfield', 'modal_dialog_textfield'), selected_color=Common.color('context_add_modal_dialog_textfield_selected', 'modal_dialog_textfield_selected'), label_color=Common.color('context_add_modal_dialog_textfield_label', 'modal_dialog_textfield_label'))
    self.add_field(self.name_field)
    self.access_key_field = DialogFieldText('Access key:', label_min=16, color=Common.color('context_add_modal_dialog_textfield', 'modal_dialog_textfield'), selected_color=Common.color('context_add_modal_dialog_textfield_selected', 'modal_dialog_textfield_selected'), label_color=Common.color('context_add_modal_dialog_textfield_label', 'modal_dialog_textfield_label'))
    self.add_field(self.access_key_field)
    self.secret_key_field = DialogFieldText('Secret key:', label_min=16, color=Common.color('context_add_modal_dialog_textfield', 'modal_dialog_textfield'), selected_color=Common.color('context_add_modal_dialog_textfield_selected', 'modal_dialog_textfield_selected'), label_color=Common.color('context_add_modal_dialog_textfield_label', 'modal_dialog_textfield_label'), password=True)
    self.add_field(self.secret_key_field)
    self.caller = caller

  def input(self, inkey):
    if not self.accepts_inputs:
      return True
    if inkey.is_sequence and inkey.name == 'KEY_ESCAPE':
      self.close()
      return True
    return super().input(inkey)

  def accept_and_close(self):
    if self.name_field.text == '':
      self.error_label.text = 'Name cannot be blank.'
      return
    if self.access_key_field.text == '':
      self.error_label.text = 'Access key cannot be blank.'
      return
    if self.secret_key_field.text == '':
      self.error_label.text = 'Secret key cannot be blank.'
      return
    try:
      sts = Common.Session.service_provider.whoami(keys={'access': self.access_key_field.text, 'secret': self.secret_key_field.text})
    except exceptions.ClientError as e:
      self.error_label.text = 'Key verification failed'
      return

    self.accepts_inputs = False

    Common.Configuration.add_or_edit_context(self.name_field.text, sts['Account'], self.access_key_field.text, self.secret_key_field.text)
    self.close()

  def close(self):
    if self.caller is not None:
      self.caller.reload_contexts()
    super().close()

class ContextList(ListControl):
  def __init__(self, parent, alignment, dimensions, *args, **kwargs):
    super().__init__(
      parent,
      alignment,
      dimensions,
      color=Common.color('context_list_generic', 'generic'),
      selection_color=Common.color('context_list_selection', 'selection'),
      title_color=Common.color('context_list_heading', 'column_title'),
      *args,
      **kwargs,
    )
    self.hotkey_display = HotkeyDisplay(self.parent, TopRightAnchor(1, 0), Dimension('33%|50', 8), self, session=Common.Session, highlight_color=Common.color('hotkey_display_title'), generic_color=Common.color('hotkey_display_value'), tag='context')
    self.add_hotkey('a', self.add_new_context, 'Add new context')
    self.add_hotkey('d', self.set_default_context, 'Set as default')
    self.add_hotkey('i', self.import_context, 'Import context')
    self.add_hotkey('KEY_ENTER', self.select_context, 'Select context')
    self.add_hotkey('\x04', self.delete_selected_context, 'Delete context')
    self.new_context_dialog = None
    self.add_column('account id', 12)
    self.add_column('default', 7)
    idx = 0

    self.reload_contexts()

  def reload_contexts(self):
    self.entries = []
    self.selected = 0
    idx = 0
    defa = 0
    for context, data in Common.Configuration['contexts'].items():
      if context == Common.Configuration['default_context']:
        defa = idx
        d = '✓'
      else:
        d = ' '
      self.add_entry(ListEntry(context, **{"account id": data["account_id"], "default": d}))

      idx += 1
    if self.selected >= len(self.entries) or self.selected < 0:
      self.selected = defa

  def add_new_context(self, _):
    AddContextDialog(self.parent, CenterAnchor(0, 0), Dimension('80%|40', '20'), caller=self, weight=-500)

  def import_context(self, _):
    ImportContextDialog(self.parent, CenterAnchor(0, 0), Dimension('80%|40', '20'), caller=self, weight=-500)

  def set_default_context(self, _):
    if self.selection is not None:
      Common.Configuration['default_context'] = self.selection.name
      Common.Configuration.write_config()
      self.reload_contexts()

  def select_context(self, _):
    if self.selection is not None:
      Common.Session.context = self.selection.name

  def delete_selected_context(self, _):
    if self.selection is not None:
      DeleteContextDialog(self.parent, CenterAnchor(0, 0), Dimension('80%|40', 10), name=self.selection.name, caller=self, weight=-500)



import re
from .common import Common, DefaultAnchor, DefaultDimension, DefaultBorder
from .termui.alignment import TopLeftAnchor, TopRightAnchor, CenterAnchor, Dimension
from .termui.dialog import DialogControl, DialogFieldLabel, DialogFieldText
from .termui.list_control import ListControl, ListEntry
from .termui.common import Commons
from .termui.control import Control, Border
from .info import HotkeyDisplay
from pathlib import Path

class CommanderOptionsLister(ListControl):
  title = "Commands"

  @classmethod
  def opener(cls, **kwargs):
    l = cls(
      Common.Session.ui.top_block,
      DefaultAnchor,
      DefaultDimension,
      weight=0,
      **kwargs
    )
    l.border=DefaultBorder("help", "Commands", None)
    return [l, l.hotkey_display]

  def __init__(self, parent, alignment, dimensions, *args, **kwargs):
    super().__init__(
      parent,
      alignment,
      dimensions,
      color=Common.color('help_list_generic', 'generic'),
      selection_color=Common.color('help_list_selection', 'selection'),
      title_color=Common.color('help_list_heading', 'column_title'),
      *args,
      **kwargs,
    )
    self.phony = {'ctx': 'AWS Contexts', 'context': 'AWS Contexts', 'region': 'AWS Region', 'ssh': 'SSH Keys'}
    self.hotkey_display = HotkeyDisplay(self.parent, TopRightAnchor(1, 0), Dimension('33%|50', 8), self, session=Common.Session, highlight_color=Common.color('hotkey_display_title'), generic_color=Common.color('hotkey_display_value'))
    self.add_hotkey('KEY_ENTER', self.select_and_close, 'Open')
    self.column_titles = {}
    self.column_order = []
    self.add_column('command', 12)
    self.add_column('resource', 48)
    self.reload()

  def reload(self, move=True):
    self.entries = []
    for cmd in Common.Session.commander_options:
      opt = Common.Session.commander_options[cmd]
      if hasattr(opt, '__self__'):
        le = ListEntry(cmd, command=cmd, resource=opt.__self__.title)
      elif cmd in self.phony:
        le = ListEntry(cmd, command=cmd, resource=self.phony[cmd])
      else:
        le = ListEntry(cmd, command=cmd, resource='')
      le.controller_data['fn'] = opt
      self.entries.append(le)
    self.entries.sort(key=lambda x: x['command'])

  def select_and_close(self, _):
    if self.selection is not None:
      Common.Session.replace_frame(self.selection.controller_data['fn']())



from .control import Control
from .common import Commons
from .color import ColorBlack, ColorGold, ColorBlackOnGold, ColorBlackOnOrange, ColorWhite
import datetime

class ListControl(Control):
  def __init__(self, parent, alignment, dimensions, color=ColorGold, selection_color=ColorBlackOnGold, title_color=ColorBlackOnOrange, *args, **kwargs):
    super().__init__(parent, alignment, dimensions, *args, **kwargs)
    self.entries = []
    self.hotkeys = {'KEY_DOWN': self.list_down, 'KEY_UP': self.list_up, 'KEY_PGUP': self.pageup, 'KEY_PGDOWN': self.pagedown, 'KEY_END': self.end, 'KEY_HOME': self.home}
    self.tooltips = {}
    self.color = color
    self.selection_color = selection_color
    self.title_color = title_color
    self.selected = 0
    self.column_titles = {"name": 30}
    self.column_order = ["name"]
    self.calculated = 0
    self._filter = None
    self.thread_share['clear'] = False
    self._cache = None
    self.thread_share['new_entries'] = []
    self.thread_share['finalize'] = False
    self.top = 0

  @property
  def filter(self):
    return self._filter

  @filter.setter
  def filter(self, value):
    Commons.UIInstance.dirty = True
    sel = self.selection
    self._filter = value
    self.top = 0
    self.selected = 0
    self._cache = None

  def add_entry(self, entry):
    self.entries.append(entry)
    self.sort()
    self._cache = None
    Commons.UIInstance.dirty = True

  def add_hotkey(self, hotkey, action, tooltip=None):
    self.hotkeys[hotkey] = action
    if tooltip is not None:
      self.tooltips[hotkey] = tooltip
    Commons.UIInstance.dirty = True

  def input(self, key):
    inkey = str(key)
    if key.is_sequence:
      inkey = key.name
    else:
      inkey = inkey.lower()
    if inkey in self.hotkeys.keys():
      self.hotkeys[inkey](self)
      Commons.UIInstance.dirty = True
      return True
    return False

  @property
  def filtered(self):
    if self._cache is None:
      self._cache = [e for e in self.entries if e.matches_filter(self._filter)]
    return self._cache

  def list_down(self, *args):
    if self.selected < len(self.filtered) - 1:
      self.selected += 1
    rows = self.rows
    if self.selected >= self.top + rows:
      self.top = max(0, self.selected - rows + 1)

  def list_up(self, *args):
    if self.selected > 0:
      self.selected -= 1
    if self.selected < self.top:
      self.top = self.selected

  def pagedown(self, *args):
    self.selected += self.rows
    self.selected = min(len(self.filtered) - 1, self.selected)
    rows = self.rows
    if self.selected >= self.top + rows:
      self.top = max(0, self.selected - rows + 1)

  def pageup(self, *args):
    self.selected -= self.rows
    self.selected = max(0, self.selected)
    if self.selected < self.top:
      self.top = self.selected

  def home(self, *args):
    self.selected = 0
    self.top = 0

  def end(self, *args):
    self.selected = len(self.filtered) - 1
    self.top = max(0, self.selected - self.rows + 1)

  @property
  def selection(self):
    if len(self.filtered) > self.selected:
      return self.filtered[self.selected]
    return None

  @property
  def rows(self):
    c = self.corners()
    y = c[1][0] + (0 if self.border is None else 1)
    y1 = c[1][1] - (0 if self.border is None else 1)
    return y1 - y

  def add_column(self, column, min_size=0, index=None):
    if index is None:
      self.column_order.append(column)
    else:
      self.column_order.insert(index, column)
    self.column_titles[column] = min_size
    self.calculated = 0
    Commons.UIInstance.dirty = True

  def before_paint_critical(self):
    if self.thread_share['clear']:
      Commons.UIInstance.dirty = True
      self.entries = []
      self.top = 0
      self.thread_share['clear'] = False
    if len(self.thread_share['new_entries']) > 0:
      Commons.UIInstance.dirty = True
      self.handle_new_entries_critical(self.thread_share['new_entries'])
      self._cache = None
      self.thread_share['new_entries'] = []
    if self.thread_share['finalize']:
      self.handle_finalization_critical()
      self.thread_share['finalize'] = False

  def handle_new_entries_critical(self, entries):
    self.entries.extend(entries)
    self.sort()

  def handle_finalization_critical(self):
    pass

  def before_paint(self):
    super().before_paint()
    self.mutex.acquire()
    try:
      self.before_paint_critical()
    finally:
      self.mutex.release()
    #if self.selected >= len(self.entries):
    #  self.selected = len(self.entries) - 1
    #if self.selected < 0:
    #  self.selected = 0

  def sort(self):
    pass

  def paint(self):
    win = self.w_in
    if win != self.calculated:
      vals = 0
      for k, v in self.column_titles.items():
        vals += v
      if vals > win:
        ratio = float(vals) / float(win)
        tot = 0
        for k in self.column_titles.keys():
          self.column_titles[k] = int(float(self.column_titles[k]) / ratio)
          tot += self.column_titles[k]
        rem = win - tot
        if 'name' in self.column_titles:
          self.column_titles["name"] += rem
        else:
          self.column_titles[self.column_order[0]] += rem
      elif vals < win:
        diff = win - vals
        part = int(diff / len(self.column_titles))
        rem = diff - (part * len(self.column_titles))
        for k in self.column_titles.keys():
          self.column_titles[k] += part
        if 'name' in self.column_titles:
          self.column_titles["name"] += rem
        else:
          self.column_titles[self.column_order[0]] += rem
      self.calculated = win

    super().paint()
    c = self.corners()
    y = c[1][0] + (0 if self.border is None else 1)
    x = c[0][0] + (0 if self.border is None else 1)
    now = datetime.datetime.now()
    rows = self.rows
    for col in self.column_order:
      text = col
      m = self.column_titles[col]
      if len(col) > m:
        text = col[:m]
      elif len(col) < m:
        text = col + (int(m - len(col)) * ' ')
      Commons.UIInstance.print(text.upper(), xy=(x,y), color=self.title_color)
      x += m
    y += 1
    for i in range(self.top, min(len(self.filtered), self.top + rows)):
      item = self.filtered[i]
      x = c[0][0] + (0 if self.border is None else 1)
      for col in self.column_order:
        text = item.columns[col]
        m = self.column_titles[col]
        if len(text) > m:
          text = text[:m]
        elif len(text) < m:
          text = text + ((m - len(text)) * ' ')
        if i == self.selected:
          if hasattr(self, 'update_selection_color') and now - item.updated < datetime.timedelta(seconds=3):
            color = self.update_selection_color
          else:
            color = self.selection_color
        else:
          if hasattr(self, 'update_color') and now - item.updated < datetime.timedelta(seconds=3):
            color = self.update_color
          else:
            color = self.color
        Commons.UIInstance.print(text, xy=(x,y), color=color)
        x += m
      y += 1

class ListEntry:
  def __init__(self, name, controller_data=None, **kwargs):
    self.name = name
    self.columns = {"name" : name}
    self.columns.update({k:str(v) for (k, v) in kwargs.items()})
    if controller_data is None:
      self.controller_data = {}
    else:
      self.controller_data = controller_data
    self.updated = datetime.datetime.now()

  def __getitem__(self, item):
    return self.columns[item]

  def __setitem__(self, item, value):
    self.columns['item'] = value
    Commons.UIInstance.dirty = True

  def matches_filter(self, filter):
    if filter is None:
      return True
    f = filter.lower()
    for k, column in self.columns.items():
      if f in column.lower():
        return True
    return False

  def mutate(self, other):
    up = False
    if self.name != other.name or len(self.columns) != len(other.columns):
      up = True
      Commons.UIInstance.log('Up because name or col len')
    else:
      for col in self.columns:
        if col not in other.columns or self.columns[col] != other.columns[col]:
          Commons.UIInstance.log('Up because col {0}'.format(col))
          up = True
          break
    if not up:
      return
    self.updated = datetime.datetime.now()
    self.name = other.name
    self.columns = other.columns.copy()
    self.controller_data = other.controller_data




from .info import InfoDisplay, NeutralDialog, HotkeyDisplay
from .termui.ui import UI, ControlCodes
from .termui.dialog import DialogFieldLabel, DialogControl
from .termui.alignment import TopLeftAnchor, BottomLeftAnchor, Dimension
import time
import subprocess
import tempfile
import os
import jq

class Session:
  def jq(self, stmt):
    if stmt not in self.jqc:
      self.jqc[stmt] = jq.compile(stmt)
    return self.jqc[stmt]

  def __init__(self, config, highlight_color, generic_color):
    self.jqc = {}
    self.ui = UI()
    self.config = config
    self.context_update_hooks = []
    self.info_display = InfoDisplay(
      self.ui.top_block,
      TopLeftAnchor(1, 0),
      Dimension('50%', 8),
      highlight_color=highlight_color,
      generic_color=generic_color,
      info=['Context', 'Region', 'UserId', 'Account', 'SSH Key', 'Default SSH username'],
      weight=-10
    )
    self.global_hotkey_tooltips = {':': 'Command palette', '/': 'Filter / search', ControlCodes.C: 'Exit'}
    self.message_display = NeutralDialog(
      self.ui.top_block,
      BottomLeftAnchor(0, 0),
      Dimension('100%', 3),
      confirm_text = None,
      cancel_text = None,
      color=generic_color,
      weight=-10,
    )
    self._message_label = DialogFieldLabel('')
    self.message_display.add_field(self._message_label)
    self._context = None
    self.context = config['default_context']
    self.region = config['default_region']
    self.ssh_key = config['default_ssh_key']
    self.stack = []
    self.stack_frame = []
    self.resource_main = None
    self.service_provider = None
    self.message_time = 0
    self.last_tick = time.time()
    self.ui.tickers.append(self.tick)

    self.filterer = None
    self.commander = None
    self.commander_options = {}

  def set_message(self, text, color):
    self._message_label.texts = [(text, color)]
    self.message_time = 5.0
    self.last_tick = time.time()
    self.ui.dirty = True

  def replace_frame(self, new_frame, drop_stack=True):
    for elem in self.stack_frame:
      elem.parent.remove_block(elem)
    if drop_stack:
      self.stack = []
    self.resource_main = new_frame[0]
    self.stack_frame = new_frame[:]
    self.ui.dirty = True

  def push_frame(self, new_frame):
    self.stack.append(self.stack_frame[:])
    self.replace_frame(new_frame, drop_stack=False)
    if hasattr(new_frame[0], 'add_hotkey'):
      if 'KEY_ESCAPE' not in new_frame[0].tooltips:
        new_frame[0].add_hotkey('KEY_ESCAPE', self.pop_frame, 'Back')
    self.ui.dirty = True

  def pop_frame(self, *args):
    for elem in self.stack_frame:
      if hasattr(elem, 'on_close'):
        elem.on_close()
    if len(self.stack) > 0:
      self.replace_frame(self.stack.pop(), drop_stack=False)
    for elem in self.stack_frame:
      elem.reparent()
    self.ui.dirty = True

  def extend_frame(self, control):
    self.stack_frame.append(control)
    self.ui.dirty = True

  def remove_from_frame(self, control):
    if hasattr(control, 'on_close'):
      control.on_close()
    self.stack_frame.remove(control)
    self.ui.dirty = True

  def tick(self):
    delta = time.time() - self.last_tick
    self.last_tick += delta
    if self.message_time > 0:
      self.message_time -= delta
      if self.message_time <= 0:
        self._message_label.texts = []
    if hasattr(self.resource_main, 'auto_refresh'):
      self.resource_main.auto_refresh()

  @property
  def context(self):
    return self._context

  @context.setter
  def context(self, value):
    if self._context == value:
      return
    self._context = value
    self.info_display['Context'] = value
    for elem in self.context_update_hooks:
      elem()

  @property
  def region(self):
    return self._region

  @region.setter
  def region(self, value):
    self._region = value
    self.info_display['Region'] = value

  @property
  def ssh_key(self):
    return self._ssh_key

  @ssh_key.setter
  def ssh_key(self, value):
    self._ssh_key = value
    self.info_display['SSH Key'] = value
    self.info_display['Default SSH username'] = self.config['default_ssh_usernames'][value] if value in self.config['default_ssh_usernames'] else ''

  def get_keypair_association(self, keypair_id):
    if keypair_id in self.config['keypair_associations']:
      return self.config['keypair_associations'][keypair_id]
    return ''

  def set_keypair_association(self, keypair_id, key_name):
    self.config['keypair_associations'][keypair_id] = key_name
    self.config.write_config()

  def textedit(self, value):
    editor = self.config['editor_command']
    temp = tempfile.NamedTemporaryFile('w', delete=False)
    tf = temp.name
    try:
      temp.write(value)
      temp.close()
      self.ui.unraw(subprocess.run, ['bash', '-c', editor.format(tf)])
      with open(tf, 'r') as temp:
        return temp.read()
    finally:
      os.unlink(temp.name)



from .control import Control
from .common import Commons
from .color import Color, ColorBlack, ColorGold, ColorBlackOnGold, ColorBlackOnOrange, ColorWhite, Palette8Bit
import pyperclip
from pygments.lexers import JsonLexer
from pygments.token import Token
import sys

class NullHighlighter:
  def __call__(self, browser, lines):
    return [[(line, browser.color)] for line in lines]

class JsonHighlighter:
  def _get_scheme_color(self, browser, color_name):
    try:
      c = browser.scheme['colors'][color_name]
    except KeyError as e:
      print(e, file=sys.stderr)
      return browser.color
    return Color(Palette8Bit(), c['foreground'], background=c['background'])

  def __call__(self, browser, lines):
    joined = '\n'.join(lines)
    lexer = JsonLexer(tabsize=2, stripnl=False, stripall=False)
    tokens = lexer.get_tokens(joined)
    current_line = []
    result_lines = []
    for token in tokens:
      if token[0] is Token.Text:
        print('Text token with value {0}'.format(token[1]), file=sys.stderr)
        brk = token[1].split('\n')
        while len(brk) > 1:
          current_line.append((brk[0], self._get_scheme_color(browser, 'syntax_highlight_{0}'.format(str(token[0]).lower().replace('.', '_')))))
          result_lines.append(current_line)
          current_line = []
          brk = brk[1:]
        current_line.append((brk[0], self._get_scheme_color(browser, 'syntax_highlight_{0}'.format(str(token[0]).lower().replace('.', '_')))))
      else:
        current_line.append((token[1], self._get_scheme_color(browser, 'syntax_highlight_{0}'.format(str(token[0]).lower().replace('.', '_')))))
    result_lines.append(current_line)
    return result_lines

class TextBrowser(Control):
  def __init__(self, parent, alignment, dimensions, color=ColorGold, filtered_color=ColorBlackOnGold, scheme=None, syntax_highlighting=False, *args, **kwargs):
    super().__init__(parent, alignment, dimensions, *args, **kwargs)
    self.entries = []
    self.hotkeys = {
      'KEY_DOWN': self.scroll_down,
      'KEY_UP': self.scroll_up,
      'KEY_LEFT': self.scroll_left,
      'KEY_RIGHT': self.scroll_right,
      'c': self.copy_contents,
      'w': self.toggle_wrap,
      'KEY_END': self.end,
      'KEY_HOME': self.home,
      'KEY_PGUP': self.pgup,
      'KEY_PGDOWN': self.pgdown,
    }
    self.tooltips = {'c': 'Copy', 'w': 'Toggle Wrap'}
    self.scheme = scheme
    self.color = color
    self.filtered_color = filtered_color
    self.lines = []
    self.display_lines = []
    self._prefilter_lines = []
    self.top = 0
    self.left = 0
    self._filter_line = 0
    self.wrap = False
    self._filter = None
    self.filter_positions = {}
    self._current_filter_match = []
    self.syntax_highlighter = JsonHighlighter() if syntax_highlighting else NullHighlighter()


  def _raw(self, line_part):
    if isinstance(line_part, tuple):
      return line_part[0]
    return line_part

  def rawlines(self):
    return [self.raw(i) for i in range(len(self.lines))]

  def raw(self, line=None):
    if line is None:
      return '\n'.join(self.rawlines())
    else:
      return ''.join([self._raw(self.lines[line][i]) for i in range(len(self.lines[line]))])

  def add_text(self, text):
    add = text.split('\n')
    self.lines.extend(add)
    print('Lines is now: \n{0}'.format(self.lines), file=sys.stderr)
    self._prefilter_lines.extend(add)
    self.display_lines = self.syntax_highlighter(self, self.lines)
    print('Display lines is now: \n{0}'.format(self.display_lines), file=sys.stderr)
    Commons.UIInstance.dirty = True

  def clear(self):
    Commons.UIInstance.dirty = True
    self.lines = []
    self.filter_positions = {}
    self._prefilter_lines = []
    self.filter = None
    self.top = 0
    self.left = 0

  @property
  def filter(self):
    return self._filter

  @filter.setter
  def filter(self, value):
    Commons.UIInstance.dirty = True
    refilter = False
    if self._filter != value:
      self._filter_line = -1
      self.filter_positions = {}
      self._current_filter_match = []
      refilter = True

    self._filter = value

    if value is None or value == '':
      return

    if refilter:
      for i in range(len(self.lines)):
        rawline = self.raw(i)
        ml = rawline.split(self._filter)
        pos = 0
        for j in range(1, len(ml)):
          pos += len(ml[j-1])
          if i not in self.filter_positions:
            self.filter_positions[i] = []
          self.filter_positions[i].append(pos)
        if len(ml) > 1:
          self._current_filter_match.append(i)

    for elem in self._current_filter_match:
      if elem > self._filter_line:
        self._filter_line = elem
        self.top = elem
        return
    if len(self._current_filter_match) > 0:
      self.top = self._current_filter_match[0]
      self._filter_line = self.top

  def add_hotkey(self, hotkey, action, tooltip=None):
    self.hotkeys[hotkey] = action
    if tooltip is not None:
      self.tooltips[hotkey] = tooltip
    Commons.UIInstance.dirty = True

  def input(self, key):
    inkey = str(key)
    if key.is_sequence:
      inkey = key.name
    else:
      inkey = inkey.lower()
    if inkey in self.hotkeys.keys():
      self.hotkeys[inkey](self)
      Commons.UIInstance.dirty = True
      return True
    return False

  def toggle_wrap(self, *args):
    self.wrap = not self.wrap
    if self.wrap:
      self.left = 0

  def scroll_down(self, *args):
    if self.top < len(self.lines) - 1:
      self.top += 1

  def scroll_up(self, *args):
    if self.top > 0:
      self.top -= 1

  def scroll_right(self, *args):
    if self.wrap:
      return
    self.left += 1

  def scroll_left(self, *args):
    if self.wrap:
      return
    if self.left > 0:
      self.left -= 1

  def end(self, *args):
    c = self.corners()
    y0 = c[1][0] + (0 if self.border is None else 1)
    y1 = c[1][1] - (0 if self.border is None else 1)
    h = y1 - y0 + 1
    limit = len(self.lines) - h if not self.wrap else len(self.lines) - 1
    self.top = max(limit, 0)

  def home(self, *args):
    self.top = 0

  def pgup(self, *args):
    c = self.corners()
    y0 = c[1][0] + (0 if self.border is None else 1)
    y1 = c[1][1] - (0 if self.border is None else 1)
    h = y1 - y0 + 1
    self.top = max(self.top - h, 0)

  def pgdown(self, *args):
    c = self.corners()
    y0 = c[1][0] + (0 if self.border is None else 1)
    y1 = c[1][1] - (0 if self.border is None else 1)
    h = y1 - y0 + 1
    limit = len(self.lines) - h if not self.wrap else len(self.lines) - 1
    self.top = max(min(limit, self.top + h), 0)

  def copy_contents(self, *args):
    pyperclip.copy(self.raw())

  def paint(self):
    super().paint()
    c = self.corners()
    y = c[1][0] + (0 if self.border is None else 1)
    y1 = c[1][1] - (0 if self.border is None else 1)
    x0 = c[0][0] + (0 if self.border is None else 1)
    x1 = c[0][1] - (0 if self.border is None else 1)
    for i in range(self.top, len(self.display_lines)):
      line = self.display_lines[i]
      skip_chars = self.left
      x = x0
      end = False
      pos = skip_chars
      next_filter_pos = -1
      next_filter_last_pos = -1
      nfp_idx = 0
      if i in self.filter_positions:
        #print(line, file=sys.stderr)
        #print('POSITIONS: {0}'.format(self.filter_positions[i]), file=sys.stderr)
        for p in self.filter_positions[i]:
          if p >= pos:
            next_filter_pos = p
            next_filter_last_pos = p + len(self._filter) - 1
            #print('SET FILTER POS: {0}-{1}'.format(next_filter_pos, next_filter_last_pos), file=sys.stderr)
            break
          elif p + len(self._filter) - 1 >= pos:
            next_filter_pos = p
            next_filter_last_pos = p + len(self._filter) - 1
          else:
            nfp_idx += 1
      for elem in line:
        buf = self._raw(elem)
        if skip_chars > len(buf):
          skip_chars -= len(buf)
          continue
        elif skip_chars > 0:
          buf = buf[skip_chars:]
          skip_chars = 0
        color = self.color if not isinstance(elem, tuple) else elem[1]

        while len(buf) > 0:
          space = x1 - x + 1
          text = buf[:space]
          #print('Printing {1} at x {0}'.format(x, text), file=sys.stderr)
          buf = buf[space:]
          while len(text) > 0:
            if next_filter_last_pos >= 0 and pos >= next_filter_pos and pos <= next_filter_last_pos:
              filt = next_filter_last_pos - pos + 1
              if filt >= len(text):
                #print('PRINT: {1}@{0} (center of current filter), pos {2}, len {3}'.format(x, text, pos, len(text)), file=sys.stderr)
                Commons.UIInstance.print(text, xy=(x, y), color=self.filtered_color)
                x += len(text)
                pos += len(text)
                text = ""
              else:
                #print('PRINT: {1}@{0} (end of current filter), pos {2}, len {3}'.format(x, text[:filt], pos, len(text[:filt])), file=sys.stderr)
                Commons.UIInstance.print(text[:filt], xy=(x, y), color=self.filtered_color)
                text = text[filt:]
                x += filt
                pos += filt
                nfp_idx += 1
                if len(self.filter_positions[i]) > nfp_idx:
                  next_filter_pos = self.filter_positions[i][nfp_idx]
                  next_filter_last_pos = next_filter_pos + len(self._filter) - 1
                else:
                  next_filter_pos = -1
                  next_filter_last_pos = -1
            elif next_filter_pos >= 0 and pos + len(text) > next_filter_pos:
              unfilt = next_filter_pos - pos
              if unfilt > 0:
                #print('PRINT: {1}@{0} (beginning of filter), pos {2}, len {3}'.format(x, text[:unfilt], pos, len(text[:unfilt])), file=sys.stderr)
                Commons.UIInstance.print(text[:unfilt], xy=(x, y), color=color)
                text = text[unfilt:]
                x += unfilt
                pos += unfilt
              filt = next_filter_last_pos - next_filter_pos + 1
              if filt >= len(text):
                #print('PRINT: {1}@{0} (middle of small filter), pos {2}, len {3}'.format(x, text, pos, len(text)), file=sys.stderr)
                Commons.UIInstance.print(text, xy=(x, y), color=self.filtered_color)
                x += len(text)
                pos += len(text)
                text = ""
              else:
                #print('PRINT: {1}@{0} (end of small filter), pos {2}, len {3}'.format(x, text[:filt], pos, len(text[:filt])), file=sys.stderr)
                Commons.UIInstance.print(text[:filt], xy=(x, y), color=self.filtered_color)
                text = text[filt:]
                x += filt
                pos += filt
                nfp_idx += 1
                if len(self.filter_positions[i]) > nfp_idx:
                  next_filter_pos = self.filter_positions[i][nfp_idx]
                  next_filter_last_pos = next_filter_pos + len(self._filter) - 1
                else:
                  next_filter_pos = -1
                  next_filter_last_pos = -1
            else:
              #print('PRINT: {1}@{0} (no filter active), pos {2}, len {3}'.format(x, text, pos, len(text)), file=sys.stderr)
              Commons.UIInstance.print(text, xy=(x, y), color=color)
              pos += len(text)
              x += len(text)
              text = ""
          if x > x1:
            y += 1
            x = x0
            if not self.wrap:
              end = True

          if end or y > y1:
            break
      if not end:
        y += 1
        x = x0
      if y > y1:
        break







from .common import Commons
from .block import Block
import threading

class BorderStyle:
  def __init__(self, chars=['-', '|', '-', '-', '-', '-']):
    self.chars = chars

  @property
  def horizontal(self):
    return self.chars[0]

  @property
  def vertical(self):
    return self.chars[1]

  @property
  def TL(self):
    return self.chars[2]

  @property
  def TR(self):
    return self.chars[3]

  @property
  def BL(self):
    return self.chars[4]

  @property
  def BR(self):
    return self.chars[5]

BorderStyleContinuous = BorderStyle(['─', '│', '┌', '┐', '└', '┘'])

class Border:
  def __init__(self, style=None, color=None, title=None, title_color=None, title_info=None, title_info_color=None):
    self.style = style
    self.color = color
    self.title = title
    self.title_color = title_color
    self.title_info = title_info
    self.title_info_color = title_info_color

  def paint(self, block):
    Commons.UIInstance.log('Painting Border', level=2)
    if self.style is None or self.color is None:
      return
    corners = block.corners()
    dimensions = block.dimensions()
    for i in range(corners[1][0], corners[1][1] + 1):
      if i == corners[1][0]:
        if self.title is None or self.title_color is None:
          outer = self.style.TL + self.style.horizontal * (dimensions[0] - 1) + self.style.TR
          Commons.UIInstance.print(outer, xy=(corners[0][0], i), color=self.color)
        else:
          total_title = self.title
          if self.title_info is not None:
            total_title = '{0} ({1})'.format(self.title, self.title_info)
          cpos = int((dimensions[0] + 1) / 2)
          spos = cpos - int(len(total_title) / 2) - 1
          slen = len(total_title) + 2
          outer = self.style.TL + self.style.horizontal * spos
          l = len(outer)
          Commons.UIInstance.print(outer, xy=(corners[0][0], i), color=self.color)
          outer = ' ' + self.title + ' '
          Commons.UIInstance.print(outer, xy=(corners[0][0] + l, i), color=self.title_color, bold=True)
          l += len(outer)
          if self.title_info is not None:
            Commons.UIInstance.print('(', xy=(corners[0][0] + l, i), color=self.title_color, bold=True)
            l += 1
            Commons.UIInstance.print(self.title_info, xy=(corners[0][0] + l, i), color=self.title_color if self.title_info_color is None else self.title_info_color, bold=True)
            l += len(self.title_info)
            Commons.UIInstance.print(') ', xy=(corners[0][0] + l, i), color=self.title_color, bold=True)
            l += 2
          outer = self.style.horizontal * (dimensions[0] - 1 - spos - slen) + self.style.TR
          Commons.UIInstance.print(outer, xy=(corners[0][0] + l, i), color=self.color)
      elif i == corners[1][1]:
        outer = self.style.BL + self.style.horizontal * (dimensions[0] - 1) + self.style.BR
        Commons.UIInstance.print(outer, xy=(corners[0][0], i), color=self.color)
      else:
        outer = self.style.vertical + ' ' * (dimensions[0] - 1) + self.style.vertical
        Commons.UIInstance.print(outer, xy=(corners[0][0], i), color=self.color)

BorderNone = Border()

class Control(Block):
  def __init__(self, parent, alignment, dimensions, weight=0, tag='default', border=None, *args, **kwargs):
    Commons.UIInstance.log('Initializing Control', level=2)
    super().__init__(parent, alignment, dimensions, weight, tag, *args, **kwargs)
    self.border = border
    self.thread_share = {}
    self.mutex = threading.Lock()

  @property
  def border(self):
    return self._border

  @border.setter
  def border(self, value):
    self._border = value
    Commons.UIInstance.dirty = True

  @property
  def inner(self):
    c = self.corners()
    x0 = c[0][0] + (0 if self.border is None else 1)
    x1 = c[0][1] - (0 if self.border is None else 1)
    y0 = c[1][0] + (0 if self.border is None else 1)
    y1 = c[1][1] - (0 if self.border is None else 1)
    return ((x0, x1), (y0, y1))

  @property
  def w_in(self):
    return self.w if self.border is None else self.w - 2

  def paint(self):
    Commons.UIInstance.log('Painting Control', level=2)
    if self.border is not None:
      self.border.paint(self)
    super().paint()



from .base_control import ResourceLister, Describer, MultiLister, NoResults, GenericDescriber, DialogFieldResourceListSelector, DeleteResourceDialog, SingleRelationLister
from .common import Common, SessionAwareDialog, BaseChart
from .termui.dialog import DialogControl, DialogFieldText, DialogFieldLabel, DialogFieldButton, DialogFieldCheckbox
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes
from .ssh import SSHList
import subprocess
from pathlib import Path
import json
import jq
import botocore
import time
import datetime
from .arn import ARN

class KeyPairResourceLister(ResourceLister):
  prefix = 'keypair_list'
  title = 'EC2 Keypairs'
  command_palette = ['key', 'keypair']

  def __init__(self, *args, **kwargs):
    self.resource_key = 'ec2'
    self.list_method = 'describe_key_pairs'
    self.item_path = '.KeyPairs'
    self.column_paths = {
      'id': '.KeyPairId',
      'name': '.KeyName',
      'fingerprint': '.KeyFingerprint',
      'association': self.determine_keypair_association,
    }
    self.imported_column_sizes = {
      'id': 16,
      'name': 32,
      'fingerprint': 48,
      'association': 16,
    }
    self.describe_command = KeyPairDescriber.opener
    self.imported_column_order = ['id', 'name', 'fingerprint', 'association']
    self.primary_key = 'id'
    self.sort_column = 'id'
    super().__init__(*args, **kwargs)
    self.add_hotkey('a', self.set_keypair_association, 'Associate with SSH key')
    self.add_hotkey('c', self.create_keypair, 'Create new keypair')
    self.add_hotkey('x', self.remove_keypair_association, 'Remove association')

  def determine_keypair_association(self, entry):
    return Common.Session.get_keypair_association(entry['KeyPairId'])

  def set_keypair_association(self, _):
    if self.selection is not None:
      Common.Session.push_frame(SSHList.selector(self.ssh_selector_callback))
      Common.Session.ui.dirty = True

  def ssh_selector_callback(self, val):
    if self.selection is not None:
      Common.Session.set_keypair_association(self.selection['id'], val)
      self.refresh_data()

  def create_keypair(self, _):
    KeyPairCreateDialog.opener(self)

  def remove_keypair_association(self, _):
    if self.selection is not None:
      Common.Session.set_keypair_association(self.selection['id'], '')
      self.refresh_data()

class KeyPairCreateDialog(SessionAwareDialog):
  def __init__(self, *args, caller=None, **kwargs):
    kwargs['border'] = Border(Common.border('default'), Common.color('modal_dialog_border'), 'Create new keypair', Common.color('modal_dialog_border_title'))
    super().__init__(*args, caller=caller,  **kwargs)
    self.add_field(DialogFieldLabel('Enter keypair details'))
    self.error_label = DialogFieldLabel('', default_color=Common.color('modal_dialog_error'))
    self.add_field(self.error_label)
    self.name_field = DialogFieldText('Name:', label_min=16, color=Common.color('modal_dialog_textfield'), selected_color=Common.color('modal_dialog_textfield_selected'), label_color=Common.color('modal_dialog_textfield_label'))
    self.add_field(self.name_field)
    self.dotssh = Path.home() / '.ssh'
    self.add_field(DialogFieldLabel('Key file will be saved in {0}'.format(str(self.dotssh.resolve()))))
    self.save_as_field = DialogFieldText('Filename:', label_min=16, color=Common.color('modal_dialog_textfield'), selected_color=Common.color('modal_dialog_textfield_selected'), label_color=Common.color('modal_dialog_textfield_label'))
    self.add_field(self.save_as_field)
    self.caller = caller

  def accept_and_close(self):
    if self.name_field.text == '':
      self.error_label.text = 'Name cannot be blank.'
      return
    if self.save_as_field.text == '':
      self.error_label.text = 'Filename cannot be blank.'
      return
    if '/' in self.save_as_field.text or '\\' in self.save_as_field.text:
      self.error_label.text = 'Filename cannot contain path separators.'
      return

    try:
      resp = Common.Session.service_provider('ec2').create_key_pair(KeyName=self.name_field.text)
      data = resp['KeyMaterial']
      with (self.dotssh / self.save_as_field.text).open('w') as f:
        f.write(data)
      Common.Session.set_keypair_association(resp['KeyPairId'], self.save_as_field.text)
    except Exception as e:
      Common.Session.set_message(str(e), Common.color('message_error'))

    super().accept_and_close()

  def close(self):
    if self.caller is not None:
      self.caller.refresh_data()
    super().close()

class KeyPairDescriber(Describer):
  prefix = 'keypair_browser'
  title = 'EC2 Keypair'

  def __init__(self, parent, alignment, dimensions, entry, *args, entry_key='id', **kwargs):
    self.resource_key = 'ec2'
    self.describe_method = 'describe_key_pairs'
    self.describe_kwarg_name = 'KeyPairIds'
    self.describe_kwarg_is_list = True
    self.object_path='.KeyPairs[0]'
    super().__init__(parent, alignment, dimensions, *args, entry=entry, entry_key=entry_key, **kwargs)



from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.padding import PKCS7
import hashlib
import getpass
import os
import sys
import yaml

class Keystore:
  def __init__(self, config):
    self.keylist_file = config.path / 'keys'
    p = getpass.getpass('Enter encryption phrase for key database: ')
    sha = hashlib.sha256(p.encode('ascii'))
    self.nonce_file = config.path / 'nonce'
    if not self.nonce_file.exists():
      self.nonce = os.urandom(16)
      with self.nonce_file.open('wb') as f:
        f.write(self.nonce)
    else:
      with self.nonce_file.open('rb') as f:
        self.nonce = f.read()
    self.keylist_cipher = Cipher(AES(sha.digest()), CBC(self.nonce))

    if not self.keylist_file.exists():
      self.keys = {}
    else:
      self.parse_keylist()

  def parse_keylist(self):
    try:
      with self.keylist_file.open('rb') as f:
        data = f.read()
      dec = self.keylist_cipher.decryptor()
      yd = dec.update(data) + dec.finalize()
      unpadder = PKCS7(256).unpadder()
      yd = unpadder.update(yd) + unpadder.finalize()
      self.keys = yaml.safe_load(yd.decode('UTF-8'))
    except ValueError:
      print('Incorrect password.')
      sys.exit(1)

  def write_keylist(self):
    enc = self.keylist_cipher.encryptor()
    yd = yaml.dump(self.keys).encode('UTF-8')
    padder = PKCS7(256).padder()
    padded = padder.update(yd) + padder.finalize()
    data = enc.update(padded) + enc.finalize()
    with self.keylist_file.open('wb') as f:
      f.write(data)

  def __getitem__(self, item):
    return self.keys[item]

  def set_key(self, name, access, secret):
    self.keys[name] = {'access': access, 'secret': secret}
    self.write_keylist()

  def delete_key(self, name):
    del(self.keys[name])
    self.write_keylist()


import re
from .common import Common, DefaultAnchor, DefaultDimension, DefaultBorder
from .termui.alignment import TopLeftAnchor, TopRightAnchor, CenterAnchor, Dimension
from .termui.dialog import DialogControl, DialogFieldLabel, DialogFieldText
from .termui.list_control import ListControl, ListEntry
from .termui.common import Commons
from .termui.control import Control, Border
from .info import HotkeyDisplay
from pathlib import Path

class SSHList(ListControl):
  @classmethod
  def opener(cls, **kwargs):
    l = cls(
      Common.Session.ui.top_block,
      DefaultAnchor,
      DefaultDimension,
      weight=0,
      **kwargs
    )
    l.border=DefaultBorder("ssh", "SSH Keys", None)
    return [l, l.hotkey_display]

  @classmethod
  def selector(cls, cb, **kwargs):
    return cls.opener(**{'selector_cb': cb, **kwargs})

  def __init__(self, parent, alignment, dimensions, selector_cb=None, *args, **kwargs):
    super().__init__(
      parent,
      alignment,
      dimensions,
      color=Common.color('ssh_key_list_generic', 'generic'),
      selection_color=Common.color('ssh_key_list_selection', 'selection'),
      title_color=Common.color('ssh_key_list_heading', 'column_title'),
      *args,
      **kwargs,
    )
    self.selector_cb = selector_cb
    self.hotkey_display = HotkeyDisplay(self.parent, TopRightAnchor(1, 0), Dimension('33%|50', 8), self, session=Common.Session, highlight_color=Common.color('hotkey_display_title'), generic_color=Common.color('hotkey_display_value'))
    self.add_hotkey('d', self.set_default_ssh_key, 'Set as default')
    self.add_hotkey('e', self.set_default_ssh_username, 'Set default username')
    if selector_cb is not None:
      self.add_hotkey('KEY_ENTER', self.select_and_close, 'Select and close')
    else:
      self.add_hotkey('KEY_ENTER', self.select_ssh_key, 'Select ssh key')
    self.add_column('usage frequency', 12)
    self.add_column('has public key', 12)
    self.add_column('default username', 20)
    self.add_column('default', 8)
    self.reload()

  def reload(self, move=True):
    self.entries = []
    idx = 0
    defa = 0
    for ssh_key in sorted(self.list_ssh_keys()):
      if ssh_key == Common.Configuration['default_ssh_key']:
        defa = idx
        d = '✓'
      else:
        d = ' '
      pubkey = Path.home() / '.ssh' / '{0}.pub'.format(ssh_key)
      has_pubkey = '✓' if pubkey.exists() else ' '
      du = Common.Configuration['default_ssh_usernames'][ssh_key] if ssh_key in Common.Configuration['default_ssh_usernames'] else ''
      self.add_entry(ListEntry(ssh_key, **{'usage frequency': 0, 'has public key': has_pubkey, 'default username': du, 'default': d}))
      idx += 1
    if move:
      self.selected = defa

  def list_ssh_keys(self):
    ssh_dir = Path.home() / '.ssh'
    ret = []
    for child in ssh_dir.iterdir():
      fn = child.name
      if re.match(r'^((id_[^.]+?)|(.*?\.pem))$', fn) is not None:
        ret.append(fn)
    return ret

  def set_default_ssh_key(self, _):
    if self.selection is not None:
      Common.Configuration['default_ssh_key'] = self.selection.name
      Common.Configuration.write_config()
      for entry in self.entries:
        if entry.name != Common.Configuration['default_ssh_key']:
          entry.columns['default'] = ' '
        else:
          entry.columns['default'] = '✓'

  def select_ssh_key(self, _):
    if self.selection is not None:
      Common.Session.ssh_key = self.selection.name

  def select_and_close(self, _):
    if self.selection is not None and self.selector_cb is not None:
      self.selector_cb(self.selection['name'])
      Common.Session.pop_frame()

  def set_default_ssh_username(self, _):
    SSHDefaultUsernameDialog(self.parent, CenterAnchor(0, 0), Dimension('80%|40', '10'), key_name=self.selection['name'], caller=self, weight=-500)

class SSHDefaultUsernameDialog(DialogControl):
  def __init__(self, parent, alignment, dimensions, key_name='', caller=None, *args, **kwargs):
    self.key_name = key_name
    kwargs['ok_action'] = self.accept_and_close
    kwargs['cancel_action'] = self.close
    kwargs['border'] = Border(
      Common.border('ssh', 'default'),
      Common.color('ssh_modal_dialog_border', 'modal_dialog_border'),
      'Default SSH username',
      Common.color('ssh_modal_dialog_border_title', 'modal_dialog_border_title'),
      self.key_name,
      Common.color('ssh_modal_dialog_border_title_info', 'modal_dialog_border_title_info'),
    )
    super().__init__(parent, alignment, dimensions, *args, **kwargs)
    def_text = Common.Configuration['default_ssh_usernames'][self.key_name] if self.key_name in Common.Configuration['default_ssh_usernames'] else ''
    self.username_textfield = DialogFieldText(
      'SSH username',
      text=def_text,
      color=Common.color('ec2_ssh_modal_dialog_textfield', 'modal_dialog_textfield'),
      selected_color=Common.color('ec2_ssh_modal_dialog_textfield_selected', 'modal_dialog_textfield_selected'),
      label_color=Common.color('ec2_ssh_modal_dialog_textfield_label', 'modal_dialog_textfield_label'),
      label_min = 16,
    )
    self.add_field(self.username_textfield)
    self.highlighted = 1 if def_text != '' else 0
    self.caller = caller

  def input(self, inkey):
    if inkey.is_sequence and inkey.name == 'KEY_ESCAPE':
      self.close()
      return True
    return super().input(inkey)

  def accept_and_close(self):
    Common.Configuration['default_ssh_usernames'][self.key_name] = self.username_textfield.text
    Common.Configuration.write_config()
    Common.Session.ssh_key = Common.Session.ssh_key # don't ask
    self.caller.reload(move=False)
    self.close()

  def close(self):
    self.parent.remove_block(self)



from .base_control import ResourceLister, Describer, MultiLister, NoResults, GenericDescriber, DialogFieldResourceListSelector, DeleteResourceDialog, SingleRelationLister
from .common import Common, SessionAwareDialog, BaseChart
from .termui.dialog import DialogControl, DialogFieldText, DialogFieldLabel, DialogFieldButton, DialogFieldCheckbox
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes
from .ssh import SSHList
import subprocess
from pathlib import Path
import json
import jq
import botocore
import time
import datetime
from .arn import ARN

class TargetGroupResourceLister(ResourceLister):
  prefix = 'target_group_list'
  title = 'Target Groups'
  command_palette = ['tg', 'targetgroup']

  def title_info(self):
    if self.rule is not None:
      return self.rule['arn']
    return None

  def __init__(self, *args, **kwargs):
    self.resource_key = 'elbv2'
    self.list_method = 'describe_target_groups'
    self.item_path = '.TargetGroups'
    if 'rule_entry' in kwargs:
      self.rule = kwargs['rule_entry']
      arns = []
      raw = self.rule.controller_data
      try:
        arns = [i['TargetGroupArn'] for i in raw['Actions'][-1]['ForwardConfig']['TargetGroups']]
        self.list_kwargs = {'TargetGroupArns': arns}
      except KeyError:
        self.rule = None
        Common.Session.set_message('Rule is not configured for forwarding.', Common.color('message_error'))
    else:
      self.rule = None
    self.column_paths = {
      'name': '.TargetGroupName',
      'protocol': '.Protocol',
      'port': '.Port',
      'target type': '.TargetType',
    }
    self.imported_column_sizes = {
      'name': 30,
      'protocol': 10,
      'port': 10,
      'target type': 10,
    }
    self.hidden_columns = {
      'arn': '.TargetGroupArn',
    }
    self.describe_command = TargetGroupDescriber.opener
    #self.open_command = ListenerActionResourceLister.opener
    #self.open_selection_arg = 'listener'

    self.imported_column_order = ['name', 'protocol', 'port', 'target type']
    self.sort_column = 'name'
    self.primary_key = 'arn'
    super().__init__(*args, **kwargs)

class TargetGroupDescriber(Describer):
  prefix = 'tg_browser'
  title = 'Target Group'

  def populate_entry(self, *args, entry, **kwargs):
    super().populate_entry(*args, entry=entry, **kwargs)
    try:
      self.name = entry['name']
    except KeyError:
      self.name = self.entry_id

  def __init__(self, parent, alignment, dimensions, entry, *args, entry_key='arn', **kwargs):
    self.resource_key = 'elbv2'
    self.describe_method = 'describe_target_groups'
    self.describe_kwarg_name = 'TargetGroupArns'
    self.describe_kwarg_is_list = True
    self.object_path='.TargetGroups[0]'
    super().__init__(parent, alignment, dimensions, *args, entry=entry, entry_key=entry_key, **kwargs)

  def title_info(self):
    return self.name



from .base_control import ResourceLister, Describer, MultiLister, NoResults, GenericDescriber, DialogFieldResourceListSelector, DeleteResourceDialog, SingleRelationLister
from .common import Common, SessionAwareDialog, BaseChart
from .termui.dialog import DialogControl, DialogFieldText, DialogFieldLabel, DialogFieldButton, DialogFieldCheckbox
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes
from .ssh import SSHList
import subprocess
from pathlib import Path
import json
import jq
import botocore
import time
import datetime
from .arn import ARN

class LBResourceLister(ResourceLister):
  prefix = 'lb_list'
  title = 'Load Balancers'
  command_palette = ['lb', 'elbv2', 'loadbalancing']

  def __init__(self, *args, **kwargs):
    from .resource_listener import ListenerResourceLister
    self.resource_key = 'elbv2'
    self.list_method = 'describe_load_balancers'
    self.item_path = '.LoadBalancers'
    self.column_paths = {
      'name': '.LoadBalancerName',
      'type': '.Type',
      'scheme': self.determine_scheme,
      'hostname': '.DNSName',
    }
    self.imported_column_sizes = {
      'name': 30,
      'type': 15,
      'scheme': 8,
      'hostname': 120,
    }
    self.hidden_columns = {
      'arn': '.LoadBalancerArn',
    }
    self.describe_command = LBDescriber.opener
    self.open_command = ListenerResourceLister.opener
    self.open_selection_arg = 'lb'

    self.imported_column_order = ['name', 'type', 'scheme', 'hostname']
    self.sort_column = 'name'
    self.primary_key = 'name'
    super().__init__(*args, **kwargs)

  def determine_scheme(self, lb, *args):
    if lb['Scheme'] == 'internet-facing':
      return 'public'
    return 'private'

class LBDescriber(Describer):
  prefix = 'lb_browser'
  title = 'Load Balancer'

  def __init__(self, parent, alignment, dimensions, entry, *args, entry_key='name', **kwargs):
    self.resource_key = 'elbv2'
    self.describe_method = 'describe_load_balancers'
    self.describe_kwarg_name = 'Names'
    self.describe_kwarg_is_list = True
    self.object_path='.LoadBalancers[0]'
    super().__init__(parent, alignment, dimensions, *args, entry=entry, entry_key=entry_key, **kwargs)

  def title_info(self):
    return self.lb_name


"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

with open('README.md', 'r') as readme:
  ld = readme.read()

setup(
  name='awsc',
  url='https://github.com/baliame/awsc',

  version='0.3.0',
  description='AWS Commander',
  long_description=ld,
  long_description_content_type='text/markdown',
  author='baliame',
  author_email='akos.toth@cheppers.com',  # Optional

  classifiers=[  # Optional
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3 :: Only',
  ],

  package_dir={'': 'src'},  # Optional
  packages=find_packages(where='src'),  # Required
  python_requires='>=3.8, <4',
  install_requires=['blessed>=1.17.12',
    'gnupg>=1.3.2',
    'cryptography>=3.1',
    'pyyaml>=5.3.1',
    'boto3>=1.16.33',
    'jq>=1.1.1',
    'pyperclip>=1.8.1',
    'numpy>=1.17.4'
  ],

  entry_points={  # Optional
    'console_scripts': [
      'awsc=awsc:main',
    ],
  },
)


class ARN:
  def __init__(self, s):
    parts = s.split(':')
    if len(parts) < 6:
      raise ValueError('ARN does not contain enough components for full identification')
    self.partition = parts[1] if parts[1] != '' else 'aws'
    if parts[2] == '':
      raise ValueError('ARN is missing service identifier.')
    self.service = parts[2]
    self.region = parts[3]
    self.account_id = parts[4]
    if len(parts) == 6:
      test = parts[5].split('/')
      if len(test) > 1:
        self.resource_type = test[0]
        self.resource_id = '/'.join(test[1:])
      else:
        self.resource_id = parts[5]
        self.resource_type = ''
    else:
      self.resource_type = parts[5]
      self.resource_id = ':'.join(parts[6:])

  @property
  def resource(self):
    if self.resource_type != '':
      return '{0}/{1}'.format(self.resource_type, self.resource_id)
    return self.resource_id

  @property
  def resource_id_first(self):
    return self.resource_id.split('/')[0]

  def __str__(self):
    return 'arn:{0}:{1}:{2}:{3}:{4}'.format(self.partition, self.service, self.region, self.account_id, self.resource)



from .base_control import ResourceLister, Describer, MultiLister, NoResults, GenericDescriber, DialogFieldResourceListSelector, DeleteResourceDialog, SingleRelationLister
from .common import Common, SessionAwareDialog, BaseChart
from .termui.dialog import DialogControl, DialogFieldText, DialogFieldLabel, DialogFieldButton, DialogFieldCheckbox
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes
from .ssh import SSHList
import subprocess
from pathlib import Path
import json
import jq
import botocore
import time
import datetime
from .arn import ARN

class DBSubnetGroupResourceLister(ResourceLister):
  prefix = 'db_subnet_group_list'
  title = 'DB Subnet Groups'
  command_palette = ['dsg', 'dbsubnetgroup']

  def __init__(self, *args, **kwargs):
    from .resource_subnet import SubnetResourceLister
    self.resource_key = 'rds'
    self.list_method = 'describe_db_subnet_groups'
    self.item_path = '.DBSubnetGroups'
    self.column_paths = {
      'name': '.DBSubnetGroupName',
      'vpc': '.VpcId',
      'status': '.SubnetGroupStatus'
    }
    self.hidden_columns = {
      'arn': '.DBSubnetGroupArn',
    }
    self.imported_column_sizes = {
      'name': 30,
      'vpc': 30,
      'status': 30,
    }
    self.describe_command = DBSubnetGroupDescriber.opener
    self.open_command = SubnetResourceLister.opener
    self.open_selection_arg = 'db_subnet_group'
    self.primary_key = 'name'

    self.imported_column_order = ['name', 'vpc', 'status']
    self.sort_column = 'name'
    super().__init__(*args, **kwargs)

class DBSubnetGroupDescriber(Describer):
  prefix = 'subnet_browser'
  title = 'Subnet'

  def __init__(self, parent, alignment, dimensions, entry, *args, entry_key='name', **kwargs):
    self.resource_key = 'rds'
    self.describe_method = 'describe_db_subnet_groups'
    self.describe_kwarg_name = 'DBSubnetGroupName'
    self.describe_kwarg_is_list = True
    self.object_path='.DBSubnetGroups[0]'
    super().__init__(parent, alignment, dimensions, *args, entry=entry, entry_key=entry_key, **kwargs)



from .base_control import ResourceLister, Describer, MultiLister, NoResults, GenericDescriber, DialogFieldResourceListSelector, DeleteResourceDialog, SingleRelationLister
from .common import Common, SessionAwareDialog, BaseChart
from .termui.dialog import DialogControl, DialogFieldText, DialogFieldLabel, DialogFieldButton, DialogFieldCheckbox
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes
from .ssh import SSHList
import subprocess
from pathlib import Path
import json
import jq
import botocore
import time
import datetime
from .arn import ARN

class RouteTableResourceLister(ResourceLister):
  prefix = 'route_table_list'
  title = 'Route Tables'
  command_palette = ['rt', 'routetable']

  def title_info(self):
    if self.subnet is not None:
      return self.subnet['id']
    return None

  def __init__(self, *args, **kwargs):
    self.resource_key = 'ec2'
    self.list_method = 'describe_route_tables'
    self.item_path = '.RouteTables'
    if 'subnet' in kwargs:
      self.subnet = kwargs['subnet']
      self.list_kwargs = {
        'Filters': [{
          'Name': 'association.subnet-id',
          'Values': [
            self.subnet['id'],
          ]
        }]
      }
    else:
      self.subnet = None
    self.column_paths = {
      'id': '.RouteTableId',
      'name': self.tag_finder_generator('Name'),
      'vpc': '.VpcId',
      'subnet': self.determine_subnet_association,
    }
    self.imported_column_sizes = {
      'id': 30,
      'name': 30,
      'vpc': 30,
      'subnet': 30,
    }
    self.describe_command = RouteTableDescriber.opener
    self.open_command = RouteResourceLister.opener
    self.open_selection_arg = 'route_table'

    self.imported_column_order = ['id', 'name', 'vpc', 'subnet']
    self.sort_column = 'id'
    self.primary_key = 'id'
    super().__init__(*args, **kwargs)

  def determine_subnet_association(self, rt, *args):
    if 'Associations' not in rt or len(rt['Associations']) == 0:
      return '<none>'
    if len(rt['Associations']) > 1:
      return '<multiple>'
    if 'SubnetId' in rt['Associations'][0]:
      return rt['Associations'][0]['SubnetId']
    else:
      return '<VPC default>'

class RouteTableDescriber(Describer):
  prefix = 'route_table_browser'
  title = 'Route Table'

  def __init__(self, parent, alignment, dimensions, entry, *args, entry_key='id', **kwargs):
    self.resource_key = 'ec2'
    self.describe_method = 'describe_route_tables'
    self.describe_kwarg_name = 'RouteTableIds'
    self.describe_kwarg_is_list = True
    self.object_path='.RouteTables[0]'
    super().__init__(parent, alignment, dimensions, *args, entry=entry, entry_key=entry_key, **kwargs)

# Routes
class RouteResourceLister(ResourceLister):
  prefix = 'routee_list'
  title = 'Routes'
  command_palette = ['route']

  def title_info(self):
    if self.route_table is not None:
      return self.route_table['id']
    return None

  def __init__(self, *args, **kwargs):
    self.resource_key = 'ec2'
    self.list_method = 'describe_route_tables'
    self.item_path = '[.RouteTables[] as $rt | $rt.Routes[] as $r | $r | .RouteTableId = $rt.RouteTableId]'
    if 'route_table' in kwargs:
      self.route_table = kwargs['route_table']
      self.list_kwargs = {
        'RouteTableIds': [self.route_table['id']]
      }
    else:
      self.route_table = None
    self.column_paths = {
      'route table': '.RouteTableId',
      'gateway type': self.determine_gateway_type,
      'gateway': self.determine_gateway,
      'destination': '.DestinationCidrBlock',
      'state': '.State',
    }
    self.hidden_columns = {
      'name': self.empty,
    }
    self.imported_column_sizes = {
      'route table': 30,
      'gateway type': 20,
      'gateway': 30,
      'destination': 30,
      'state': 10,
    }

    self.imported_column_order = ['route table', 'gateway type', 'gateway', 'destination', 'state']
    self.sort_column = 'route table'
    self.primary_key = None
    super().__init__(*args, **kwargs)

    self.add_hotkey('d', self.generic_describe, 'Describe')
    self.add_hotkey('KEY_ENTER', self.generic_describe, 'Describe')

  def generic_describe(self, entry):
    if self.selection is not None:
      Common.Session.push_frame(GenericDescriber.opener(**{
        'describing': 'Route in route table {0}'.format(self.selection['route table']),
        'content': json.dumps(self.selection.controller_data, sort_keys=True, indent=2),
        'pushed': True,
      }))

  def determine_gateway_type(self, entry):
    if 'NatGatewayId' in entry:
      return 'NAT'
    elif 'InstanceId' in entry:
      return 'Instance'
    elif 'TransitGatewayId' in entry:
      return 'Transit'
    elif 'LocalGatewayId' in entry:
      return 'Local'
    elif 'CarrierGatewayId' in entry:
      return 'Carrier'
    elif 'VpcPeeringConnectionId' in entry:
      return 'VPC Peering'
    elif 'EgressOnlyInternetGatewayId' in entry:
      return 'Egress-Only'
    elif entry['GatewayId'] == 'local':
      return 'VPC-Local'
    else:
      return 'Internet'

  def determine_gateway(self, entry):
    if 'NatGatewayId' in entry:
      return entry['NatGatewayId']
    elif 'InstanceId' in entry:
      return entry['InstanceId']
    elif 'TransitGatewayId' in entry:
      return entry['TransitGatewayId']
    elif 'LocalGatewayId' in entry:
      return entry['LocalGatewayId']
    elif 'CarrierGatewayId' in entry:
      return entry['CarrierGatewayId']
    elif 'VpcPeeringConnectionId' in entry:
      return entry['VpcPeeringConnectionId']
    elif 'EgressOnlyInternetGatewayId' in entry:
      return entry['EgressOnlyInternetGatewayId']
    else:
      return entry['GatewayId']



from .base_control import ResourceLister, Describer, MultiLister, NoResults, GenericDescriber, DialogFieldResourceListSelector, DeleteResourceDialog, SingleRelationLister
from .common import Common, SessionAwareDialog, BaseChart
from .termui.dialog import DialogControl, DialogFieldText, DialogFieldLabel, DialogFieldButton, DialogFieldCheckbox
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes
from .ssh import SSHList
import subprocess
from pathlib import Path
import json
import jq
import botocore
import time
import datetime
from .arn import ARN

class R53ResourceLister(ResourceLister):
  prefix = 'r53_list'
  title = 'Route53 Hosted Zones'
  command_palette = ['r53', 'route53']

  def __init__(self, *args, **kwargs):
    self.resource_key = 'route53'
    self.list_method = 'list_hosted_zones'
    self.item_path = '.HostedZones'
    self.column_paths = {
      'id': '.Id',
      'name': '.Name',
      'records': '.ResourceRecordSetCount',
    }
    self.imported_column_sizes = {
      'id': 30,
      'name': 30,
      'records': 5,
    }
    self.describe_command = R53Describer.opener
    self.open_command = R53RecordLister.opener
    self.open_selection_arg = 'r53_entry'
    self.primary_key = 'id'

    self.imported_column_order = ['id', 'name', 'records']
    self.sort_column = 'name'
    super().__init__(*args, **kwargs)

class R53RecordLister(ResourceLister):
  prefix = 'r53_record_list'
  title = 'Route53 Records'

  def title_info(self):
    return self.r53_entry['name']

  def __init__(self, *args, r53_entry, **kwargs):
    self.r53_entry = r53_entry
    self.resource_key = 'route53'
    self.list_method = 'list_resource_record_sets'
    self.list_kwargs={'HostedZoneId': self.r53_entry['id']}
    self.item_path = '.ResourceRecordSets'
    self.column_paths = {
      'entry': '.Type',
      'name': self.determine_name,
      'records': self.determine_records,
      'ttl': '.TTL',
    }
    self.hidden_columns = {
      'hosted_zone_id': self.determine_hosted_zone_id,
    }
    self.imported_column_sizes = {
      'entry': 5,
      'name': 30,
      'records': 60,
      'ttl': 5,
    }
    self.describe_command = R53RecordDescriber.opener

    self.imported_column_order = ['entry', 'name', 'records', 'ttl']
    self.sort_column = 'name'
    self.primary_key = None
    super().__init__(*args, **kwargs)
    self.add_hotkey('e', self.edit, 'Edit')

  def determine_hosted_zone_id(self, s):
    return self.r53_entry['id']

  def determine_name(self, s):
    return s['Name'].replace('\\052', '*')

  def edit(self, _):
    if self.selection is not None:
      raw = self.selection.controller_data
      if not self.is_alias(raw):
        content = '\n'.join(record['Value'] for record in raw['ResourceRecords'])
        newcontent = Common.Session.textedit(content).strip(' \n\t')
        if content == newcontent:
          Common.Session.set_message('Input unchanged.', Common.color('message_info'))
          return
        newcontent = newcontent.split('\n')
        try:
          Common.Session.service_provider(self.resource_key).change_resource_record_sets(
            HostedZoneId=self.selection['hosted_zone_id'],
            ChangeBatch={
              'Changes': [{
                'Action': 'UPSERT',
                'ResourceRecordSet': {
                  'Name': self.selection['name'],
                  'Type': self.selection['entry'],
                  'ResourceRecords': [{'Value': line} for line in newcontent],
                  'TTL': int(self.selection['ttl']),
                },
              }]
            }
          )
        except botocore.errorfactory.InvalidInput as e:
          Common.Session.ui.log(str(e))
          Common.Session.set_message('AWS API returned error, logged.', Common.color('message_error'))
          return
          Common.Session.set_message('Entry modified, refreshing...', Common.color('message_success'))

        self.refresh_data()
      else:
        Common.Session.set_message('Cannot edit aliased records', Common.color('message_info'))

  def is_alias(self, s):
    return 'AliasTarget' in s and s['AliasTarget']['DNSName'] != ''

  def determine_records(self, s):
    if self.is_alias(s):
      return s['AliasTarget']['DNSName']
    return ','.join([record['Value'] for record in s['ResourceRecords']])

class R53Describer(Describer):
  prefix = 'r53_browser'
  title = 'Route53 Hosted Zone'

  def __init__(self, parent, alignment, dimensions, entry, *args, entry_key='id', **kwargs):
    self.resource_key = 'route53'
    self.describe_method = 'get_hosted_zone'
    self.describe_kwarg_name = 'Id'
    self.object_path='.'
    super().__init__(parent, alignment, dimensions, *args, entry=entry, entry_key=entry_key, **kwargs)

class R53RecordDescriber(Describer):
  prefix = 'r53_record_browser'
  title = 'Route53 Record'

  def populate_entry(self, *args, entry, **kwargs):
    super().populate_entry(*args, entry=entry, **kwargs)
    self.record_type = entry['entry']
    self.record_name = entry['name']

  def populate_describe_kwargs(self, *args, **kwargs):
    super().populate_describe_kwargs(*args, **kwargs)
    self.describe_kwargs['StartRecordType'] = self.record_type
    self.describe_kwargs['StartRecordName'] = self.record_name

  def __init__(self, parent, alignment, dimensions, entry, *args, entry_key='hosted_zone_id', **kwargs):
    self.resource_key = 'route53'
    self.describe_method = 'list_resource_record_sets'
    self.describe_kwarg_name = 'HostedZoneId'
    self.object_path='.ResourceRecordSets[0]'
    super().__init__(parent, alignment, dimensions, *args, entry=entry, entry_key=entry_key, **kwargs)

  def title_info(self):
    return '{0} {1}'.format(self.record_type, self.record_name)



class Commons:
  UIInstance = None
  TextfieldInputs = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890/:@#.$=_-[]{}()<>|%!+~? '
  TextfieldInputsAlphaNum = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890'



import yaml

class Scheme:
  def __init__(self, config):
    self.style_file = config.path / 'style.yaml'
    self.config = config
    if not self.style_file.exists():
      self.create_default_config()
    else:
      self.parse_config()

  def backup_and_reset(self):
    print('Restoring style scheme to defaults. Creating backup of pre-restore scheme.')
    backup = self.config.path / 'style.yaml.bak'
    self.style_file.rename(backup)
    self.create_default_config()

  def create_default_config(self):
    print('Creating first time style scheme...')
    self.style = {
      'colors': {
        'generic': {
          'foreground': 220,
          'background': 0,
        },
        'highlight': {
          'foreground': 70,
          'background': 0,
        },
        'highlight_selection': {
          'foreground': 0,
          'background': 70,
        },
        'info_display_title': {
          'foreground': 70,
          'background': 0,
        },
        'info_display_value': {
          'foreground': 220,
          'background': 0,
        },
        'hotkey_display_title': {
          'foreground': 70,
          'background': 0,
        },
        'hotkey_display_value': {
          'foreground': 220,
          'background': 0,
        },
        'modal_dialog_label_highlight': {
          'foreground': 70,
          'background': 0,
        },
        'modal_dialog_label': {
          'foreground': 220,
          'background': 0,
        },
        'generic_border': {
          'foreground': 220,
          'background': 0,
        },
        'modal_dialog_border': {
          'foreground': 220,
          'background': 0,
        },
        'border_title': {
          'foreground': 111,
          'background': 0,
        },
        'border_title_info': {
          'foreground': 70,
          'background': 0,
        },
        'modal_dialog_border_title': {
          'foreground': 111,
          'background': 0,
        },
        'modal_dialog_border_title_info': {
          'foreground': 70,
          'background': 0,
        },
        'column_title': {
          'foreground': 0,
          'background': 208,
        },
        'message_success': {
          'foreground': 70,
          'background': 0,
        },
        'message_info': {
          'foreground': 33,
          'background': 0,
        },
        'message_error': {
          'foreground': 124,
          'background': 0,
        },
        'error': {
          'foreground': 124,
          'background': 0,
        },
        'modal_dialog_error': {
          'foreground': 124,
          'background': 0,
        },
        'selection': {
          'foreground': 0,
          'background': 220,
        },
        'textfield_label': {
          'foreground': 220,
          'background': 0,
        },
        'textfield': {
          'foreground': 0,
          'background': 208,
        },
        'textfield_selection': {
          'foreground': 0,
          'background': 220,
        },
        'button': {
          'foreground': 220,
          'background': 0,
        },
        'button_selection': {
          'foreground': 0,
          'background': 220,
        },
        'context_list_generic': {
          'foreground': 220,
          'background': 0,
        },
        'context_list_selection': {
          'foreground': 0,
          'background': 220,
        },
        'context_list_border': {
          'foreground': 220,
          'background': 0,
        },
        'context_list_border_title': {
          'foreground': 111,
          'background': 0,
        },
        'context_list_heading': {
          'foreground': 0,
          'background': 208,
        },
        'search_bar_border': {
          'foreground': 220,
          'background': 0,
        },
        'search_bar_color': {
          'foreground': 220,
          'background': 0,
        },
        'search_bar_symbol_color': {
          'foreground': 70,
          'background': 0,
        },
        'search_bar_autocomplete_color': {
          'foreground': 239,
          'background': 0,
        },
        'search_bar_inactive_color': {
          'foreground': 239,
          'background': 0,
        },
        'command_bar_border': {
          'foreground': 220,
          'background': 0,
        },
        'command_bar_color': {
          'foreground': 220,
          'background': 0,
        },
        'command_bar_symbol_color': {
          'foreground': 70,
          'background': 0,
        },
        'command_bar_autocomplete_color': {
          'foreground': 239,
          'background': 0,
        },
        'command_bar_ok_color': {
          'foreground': 33,
          'background': 0,
        },
        'command_bar_error_color': {
          'foreground': 124,
          'background': 0,
        },
        'modal_dialog_textfield': {
          'foreground': 0,
          'background': 208,
        },
        'modal_dialog_textfield_selected': {
          'foreground': 0,
          'background': 220,
        },
        'modal_dialog_textfield_label': {
          'foreground': 220,
          'background': 0,
        },
        'syntax_highlight_token_text': {
          'foreground': 220,
          'background': 0,
        },
        'syntax_highlight_token_error': {
          'foreground': 196,
          'background': 0,
        },
        'syntax_highlight_token_other': {
          'foreground': 243,
          'background': 0,
        },
        'syntax_highlight_token_keyword': {
          'foreground': 183,
          'background': 0,
        },
        'syntax_highlight_token_keyword_constant': {
          'foreground': 183,
          'background': 0,
        },
        'syntax_highlight_token_keyword_declaration': {
          'foreground': 183,
          'background': 0,
        },
        'syntax_highlight_token_keyword_namespace': {
          'foreground': 183,
          'background': 0,
        },
        'syntax_highlight_token_keyword_pseudo': {
          'foreground': 183,
          'background': 0,
        },
        'syntax_highlight_token_keyword_reserved': {
          'foreground': 183,
          'background': 0,
        },
        'syntax_highlight_token_keyword_type': {
          'foreground': 183,
          'background': 0,
        },
        'syntax_highlight_token_literal': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_date': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_affix': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_backtick': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_char': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_delimiter': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_double': {
          'foreground': 154,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_escape': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_heredoc': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_interpol': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_other': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_regex': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_single': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_string_symbol': {
          'foreground': 193,
          'background': 0,
        },
        'syntax_highlight_token_literal_number': {
          'foreground': 111,
          'background': 0,
        },
        'syntax_highlight_token_literal_number_bin': {
          'foreground': 111,
          'background': 0,
        },
        'syntax_highlight_token_literal_number_float': {
          'foreground': 111,
          'background': 0,
        },
        'syntax_highlight_token_literal_number_hex': {
          'foreground': 111,
          'background': 0,
        },
        'syntax_highlight_token_literal_number_integer': {
          'foreground': 111,
          'background': 0,
        },
        'syntax_highlight_token_literal_number_integer_long': {
          'foreground': 111,
          'background': 0,
        },
        'syntax_highlight_token_literal_number_oct': {
          'foreground': 111,
          'background': 0,
        },
        'syntax_highlight_token_operator': {
          'foreground': 93,
          'background': 0,
        },
        'syntax_highlight_token_operator_word': {
          'foreground': 93,
          'background': 0,
        },
        'syntax_highlight_token_punctuation': {
          'foreground': 15,
          'background': 0,
        },
        'syntax_highlight_token_punctuation_marker': {
          'foreground': 15,
          'background': 0,
        },
        'syntax_highlight_token_name': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_tag': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_attribute': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_builtin': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_builtin_pseudo': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_class': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_constant': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_decorator': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_entity': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_exception': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_function': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_function_magic': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_label': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_namespace': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_other': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_variable': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_variable_class': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_variable_global': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_variable_instance': {
          'foreground': 64,
          'background': 0,
        },
        'syntax_highlight_token_name_variable_magic': {
          'foreground': 64,
          'background': 0,
        },
      },
      'borders': {
        'default': {
          'horizontal': '─',
          'vertical': '│',
          'TL': '┌',
          'TR': '┐',
          'BL': '└',
          'BR': '┘',
        },
        'search_bar': {
          'horizontal': '─',
          'vertical': '│',
          'TL': '┌',
          'TR': '┐',
          'BL': '└',
          'BR': '┘',
        },
        'modal': {
          'horizontal': '─',
          'vertical': '│',
          'TL': '┌',
          'TR': '┐',
          'BL': '└',
          'BR': '┘',
        },
        'resource_list': {
          'horizontal': '─',
          'vertical': '│',
          'TL': '┌',
          'TR': '┐',
          'BL': '└',
          'BR': '┘',
        },
      }
    }

    with self.style_file.open('w') as f:
      f.write(yaml.dump(self.style))

  def __getitem__(self, item):
    return self.style[item]

  def parse_config(self):
    with self.style_file.open('r') as f:
      self.style = yaml.safe_load(f.read())
